// -*-mode:c++; c-style:k&r; c-basic-offset:4;-*-
//
// Copyright 2019-2020, Julian Catchen <jcatchen@illinois.edu>
//
// This file is part of Chromonomer.
//
// Chromonomer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Chromonomer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Chromonomer.  If not, see <http://www.gnu.org/licenses/>.
//

#include "synteny.h"

int
traverse_nodes_for_synteny(list<Node> &lg,
                           ofstream &fh,
                           map<string, vector<Ortholog>> &con_synteny)
{
    map<ori_type, string> print_o;
    print_o[ori_unknown] = "Unknown";
    print_o[ori_forward] = "Forward";
    print_o[ori_reverse] = "Reverse";

    fh << "\n"
       << "Incorporating conserved syntney data to order and orient scaffolds using orthologous gene locations.\n\n";
    vector<string> anchored_scaffolds;
    set<string>    all_scaffolds;
    for (auto it = lg.begin(); it != lg.end(); it++) {
	Node &node = *it;

	for (auto i = node.scaffolds.begin(); i != node.scaffolds.end(); i++) {
	    Scaffold &scaffold = i->second;

            if (con_synteny.count(scaffold.id) > 0) {
                //
                // We only want to look at anchored scaffolds for our global chromosome
                // orientation. But we need to look at all scaffolds to reorder genes.
                //
                if (node.cMs.size() > 1)
                    anchored_scaffolds.push_back(scaffold.id);
                all_scaffolds.insert(scaffold.id);
            }
	}
    }    
    //
    // We need to know the overall orientation of the linkage group to the orthologous
    // chromosome before we can compare synteny.
    //
    orient_orthologous_chromosome(fh, anchored_scaffolds, all_scaffolds, con_synteny);
    
    for (auto it = lg.begin(); it != lg.end(); it++) {
	Node &node = *it;

        if (node.cMs.size() > 1)
            continue;
        
        fh << "Map node " << node.id() << " contains " << node.scaffolds.size() << " unordered scaffolds.\n";

        vector<string> working_list;
	for (auto i = node.scaffolds.begin(); i != node.scaffolds.end(); i++) {
	    Scaffold &scaffold = i->second;

            fh << "  "
               << scaffold.id << " "
               << scaffold.start_bp << "-"
               << scaffold.end_bp << "bp; orientation: '" << print_o[scaffold.ori];

            //
            // How many of these scaffolds have ortholog genes recorded?
            //
            if (con_synteny.count(scaffold.id) == 0) {
               fh << "'; 0 orthologous genes, unable to process further.\n";
            } else {
                working_list.push_back(scaffold.id);
                fh << "'; " << con_synteny[scaffold.id].size() << " orthologous genes.\n";
            }
	}
            
        if (working_list.size() >= 1) {
            order_scaffolds_by_synteny(fh, working_list, node, con_synteny);
        }
    }

    return 0;
}

int
orient_orthologous_chromosome(ostream &fh,
                              vector<string> &anchored_scaffolds,
                              set<string> &all_scaffolds,
                              map<string, vector<Ortholog>> &consyn)
{
    map<string, size_t> gene_cnts;
    vector<pair<string, size_t>> ordered_cnts;
    //
    // First determine which chromosome the majority of orthologs occur on.
    //
    for (auto it = all_scaffolds.begin(); it != all_scaffolds.end(); it++) {
        
        vector<Ortholog> &orths = consyn[*it];
        for (uint j = 0; j < orths.size(); j++)
            gene_cnts[orths[j].orth_chr]++;
    }
    for (auto it = gene_cnts.begin(); it != gene_cnts.end(); it++)
        ordered_cnts.push_back(*it);

    sort(ordered_cnts.begin(), ordered_cnts.end(),
         [](pair<string, size_t> a, pair<string, size_t> b) { return a.second > b.second; });

    string orth_chr = ordered_cnts.size() > 0 ? ordered_cnts.front().first : "";

    fh << "Orthologous chromosome (based on count of ortholgous genes) is: " << orth_chr << "\n";

    //
    // 1. Extract genes on the ortholgous chromosome from which to determine orientation.
    // 2. Assign an linkage-group level basepair to each gene to compute the orientation regression.
    //
    vector<Ortholog> orthologs;
    size_t running_bp = 0;
    for (auto it = anchored_scaffolds.begin(); it != anchored_scaffolds.end(); it++) {
        vector<Ortholog> &orths = consyn[*it];
        for (uint j = 0; j < orths.size(); j++)
            if (orths[j].orth_chr == orth_chr) {
                orths[j].bp = running_bp + orths[j].bp;
                orthologs.push_back(orths[j]);
            }
        running_bp = orths.back().bp > orths.front().bp ? orths.back().bp : orths.front().bp;
    }

    // for (uint i = 0; i < orthologs.size(); i++)
    //     cerr << orthologs[i].id << "\t" << orthologs[i].scaffold << "\t" << orthologs[i].bp << "\t" << orthologs[i].orth_chr << "\t" << orthologs[i].orth_bp << "\n";
    
    fh << "  " << orthologs.size() << " orthologous genes available on " << orth_chr << " to determine synteny ordering.\n";

    ori_type chr_ori = determine_scaffold_orientation_by_synteny(fh, orthologs);

    if (chr_ori == ori_forward) 
        fh << "  According to conserved synteny orthologous chromosome, " << orth_chr << " has orientation 'forward'.\n";

    else if (chr_ori == ori_reverse) {
        fh << "  According to conserved synteny orthologous chromosome, " <<  orth_chr << " has orientation 'reverse'.\n"
           << "    ->The orthologous chromosome will be reversed for syntenic scaffold ordering.\n";
        //
        // Now pull a list of all orthologs to reorder, not just those on anchored scaffolds which we used to determine orientation.
        //
        vector<Ortholog> all_orthologs;
        for (auto it = all_scaffolds.begin(); it != all_scaffolds.end(); it++) {
            vector<Ortholog> &orths = consyn[*it];
            for (uint j = 0; j < orths.size(); j++)
                if (orths[j].orth_chr == orth_chr)
                    all_orthologs.push_back(orths[j]);
        }
        sort(all_orthologs.begin(), all_orthologs.end(), [] (Ortholog a, Ortholog b) { return a.orth_bp < b.orth_bp; });
            
        //
        // Reverse the basepair coordinates of the orthologous gene positions.
        //
        reverse_orth_gene_order(all_orthologs);

        map<string, Ortholog> orth_dict;
        for (uint j = 0; j < all_orthologs.size(); j++)
            orth_dict[all_orthologs[j].id] = all_orthologs[j];
        
        //
        // Update the original set of orthologs so the change in order is persistent.
        //
        for (auto it = all_scaffolds.begin(); it != all_scaffolds.end(); it++) {
            vector<Ortholog> &orths = consyn[*it];
            for (uint j = 0; j < orths.size(); j++)
                orths[j].orth_bp = orth_dict[orths[j].id].orth_bp;
        }
        
    } else {
        fh << "  Unable to determine the orientation of " <<  orth_chr << " using conserved synteny.\n";
    }

    return 0;
}

int
order_scaffolds_by_synteny(ostream &fh,
                           vector<string> &scaffolds,
                           Node &node,
                           map<string, vector<Ortholog>> &con_synteny)
{
    map<ori_type, string> print_o;
    print_o[ori_unknown] = "Unknown";
    print_o[ori_forward] = "Forward";
    print_o[ori_reverse] = "Reverse";

    //
    // Check which ortholous chromosome(s) genes occur on, keep genes from the most frequent chromosome.
    //
    map<string, size_t> orth_chrs;
    for (uint i = 0; i < scaffolds.size(); i++) {
        vector<Ortholog> &genes = con_synteny[scaffolds[i]];
        for (uint j = 0; j < genes.size(); j++)
            orth_chrs[genes[j].orth_chr]++;
    }

    vector<pair<string, size_t>> orth_chr_pairs;
    for (auto it = orth_chrs.begin(); it != orth_chrs.end(); it++)
        orth_chr_pairs.push_back(make_pair(it->first, it->second));

    sort(orth_chr_pairs.begin(), orth_chr_pairs.end(),
         [](pair<string, size_t> a, pair<string, size_t> b) {
             return a.second > b.second;
         });

    string orth_chr = orth_chr_pairs[0].first;

    size_t discarded_genes = 0;
    string discarded_chrs  = "";
    for (uint i = 1; i < orth_chr_pairs.size(); i++) {
        discarded_genes += orth_chr_pairs[i].second;
        discarded_chrs  += orth_chr_pairs[i].first;
        if (i <= orth_chr_pairs.size() - 2)
            discarded_chrs += ", ";
    }
    
    if (discarded_genes == 0)
        fh << "  Keeping " << orth_chr_pairs[0].second << " genes from orthologous chromosome '" << orth_chr
           << "'; discarding " << discarded_genes << " genes.\n";
    else
        fh << "  Keeping " << orth_chr_pairs[0].second << " genes from orthologous chromosome '" << orth_chr
           << "'; discarding " << discarded_genes << " genes from " << discarded_chrs << ".\n";

    //
    // Discard genes not on the selected orthologous chromosome.
    //
    vector<Ortholog> orths;
    for (uint i = 0; i < scaffolds.size(); i++) {
        vector<Ortholog> &genes = con_synteny[scaffolds[i]];
        for (uint j = 0; j < genes.size(); j++)
            if (genes[j].orth_chr == orth_chr)
                orths.push_back(genes[j]);
    }

    //
    // Sort the remaining genes according to their ortholgous position.
    //
    sort(orths.begin(), orths.end(), [](Ortholog a, Ortholog b) { return a.orth_bp < b.orth_bp; });

    //
    // Break the genes back down by scaffold now.
    //
    map<string, vector<Ortholog>> scaf_genes;
    for (uint i = 0; i < orths.size(); i++)
        scaf_genes[orths[i].scaffold].push_back(orths[i]);

    //
    // Determine scaffold orientation.
    //
    map<string, ori_type> scaf_ori;
    for (auto it = scaf_genes.begin(); it != scaf_genes.end(); it++) {
        fh << "  " << it->first << ":\n";

        //
        // Remove genes that are anomously too far away from the other genes (or orientation may not be correctly determined).
        //
        remove_anomolously_located_genes(fh, it->second);

        ori_type o = determine_scaffold_orientation_by_synteny(fh, it->second);
        scaf_ori[it->first] = o;

        if (o == ori_forward) 
            fh << "    According to conserved synteny, " << it->first << " has orientation 'forward'.\n";
        else if (o == ori_reverse)
            fh << "    According to conserved synteny, " << it->first << " has orientation 'reverse'.\n";
        else {
            fh << "    Unable to determine the orientation of " << it->first << " using conserved synteny.\n";
        }
    }

    if (scaffolds.size() > 1) {
        //
        // Prune out-of-order genes.
        //
        for (auto it = scaf_genes.begin(); it != scaf_genes.end(); it++)
            if (scaf_ori[it->first] != ori_unknown) { 
                fh << "  Finding an ordered set of orthologous genes on " << it->first << ".\n";
                prune_outoforder_synteny_genes(fh, scaf_ori[it->first], it->second);
            }
    }

    //
    // Determine final ordering.
    //
    orths.clear();
    for (auto it = scaf_genes.begin(); it != scaf_genes.end(); it++)
        orths.insert(orths.end(), it->second.begin(), it->second.end());
    sort(orths.begin(), orths.end(), [](Ortholog a, Ortholog b) { return a.orth_bp < b.orth_bp; });

    set<string> seen_scaffolds;
    scaffolds.clear();
    for (uint i = 0; i < orths.size(); i++)
        if (seen_scaffolds.count(orths[i].scaffold) == 0) {
            seen_scaffolds.insert(orths[i].scaffold);
            scaffolds.push_back(orths[i].scaffold);
        }

    fh << "Final scaffold ordering, according to conserved synteny: ";
    for (uint i = 0; i < scaffolds.size(); i++) {
        fh << scaffolds[i] << " [" << print_o[scaf_ori[scaffolds[i]]] << "]";
        i < scaffolds.size() - 1 ? fh << ", " : fh << " ";
    }
    fh << "\n\n";

    //
    // Record the changes in the Node object.
    //
    for (uint i = 0; i < scaffolds.size(); i++)
        node.scaffolds[scaffolds[i]].ori = scaf_ori[scaffolds[i]];
        
    node.set_ordered_scaffolds(scaffolds);
    
    return 0;
}

int
reverse_orth_gene_order(vector<Ortholog> &genes)
{
    if (genes.size() <= 1)
        return 0;

    size_t max_bp = genes.back().orth_bp;
    // cerr<< "max_bp: " << max_bp << "\n";
    vector<Ortholog> reordered_genes;

    for (int i = (int) genes.size() - 1; i >= 0; i--) {
        // cerr << "  i: " << i << " old bp: " << genes[i].orth_bp;
        genes[i].orth_bp = 1 + (max_bp - genes[i].orth_bp);
        // cerr << ", new bp: " << genes[i].orth_bp << "\n";
        reordered_genes.push_back(genes[i]);
    }
    genes.clear();

    for (uint i = 0; i < reordered_genes.size(); i++) {
        genes.push_back(reordered_genes[i]);
    }
    
    return 0;
}

int
remove_anomolously_located_genes(ostream &fh, vector<Ortholog> &genes)
{
    if (genes.size() <= 1)
        return 0;

    //
    // If orthologous genes are too far away from the main syntenic cluster, the
    // linear regression to determine orientation will fail.
    //
    // Calculate a trimmed mean to remove outlier genes. Based on:
    //  J. Bednar and T. Watt. (1984) Alpha-trimmed means and their relationship to median filters.
    //
    double alpha = 0.10;

    // for (uint i = 0; i < genes.size(); i++)
    //     cerr << "orth bp: " << genes[i].orth_bp << " scaf bp: " << genes[i].bp << "\n";

    //
    // Determine the distances between sorted ortholgous genes.
    //
    vector<double> dists;
    for (uint i = 1; i < genes.size(); i++)
        dists.push_back(genes[i].orth_bp - genes[i-1].orth_bp);

    sort(dists.begin(), dists.end());
    
    //
    // We want to trim 10% off either end of the distribution of distances.
    //
    size_t trim_lim = uint(dists.size() * alpha + 0.5);

    //
    // Not enough genes to trim any away.
    //
    if (trim_lim == 0)
        return 0;
    
    size_t max_dist = dists[dists.size() - trim_lim];

    //
    // Start at the midpoint gene, exclude any genes whose distance is more than max_dist.
    //
    vector<Ortholog> trimmed_genes;

    size_t start_up, start_down;
    if (genes.size() % 2 == 1) {
        start_up   = size_t(ceil(genes.size() / 2.0));
        start_down = size_t(floor(genes.size() / 2.0));
    } else {
        start_up   = genes.size() / 2 + 1;
        start_down = genes.size() / 2;
    }

    size_t dist;
    size_t i = start_down;
    while (i >= 1) {
        dist = genes[i].orth_bp - genes[i-1].orth_bp;

        if (dist <= max_dist) {
            trimmed_genes.push_back(genes[i]);

            if (i == 1)
                trimmed_genes.push_back(genes[0]);
        } else {
            break;
        }
        i--;
    }

    i = start_up;
    while (i < genes.size() - 1) {
        dist = genes[i + 1].orth_bp - genes[i].orth_bp;

        if (dist <= max_dist) {
            trimmed_genes.push_back(genes[i]);

            if (i == genes.size() - 2)
                trimmed_genes.push_back(genes[genes.size() - 1]);
        } else {
            break;
        }
        i++;
    }

    fh << "    Trimming " << genes.size() - trimmed_genes.size()
       << " gene(s) from either end of the syntney cluster [trim limit: " << trim_lim << ", max dist: " << max_dist << "bp].\n";
    
    genes.clear();
    for (uint i = 0; i < trimmed_genes.size(); i++)
        genes.push_back(trimmed_genes[i]);

    //
    // Sort the remaining genes according to their ortholgous position.
    //
    sort(genes.begin(), genes.end(), [](Ortholog a, Ortholog b) { return a.orth_bp < b.orth_bp; });

    return 0;
}

ori_type
determine_scaffold_orientation_by_synteny(ostream &fh, vector<Ortholog> &genes)
{
    //
    // We will caclulate the linear regression line if plotting ortholog basepair positions
    // on the X-axis versus scaffold basepair positions on the Y-axis. If the slope is positive,
    // the orientation is 'forward'; if the slope is negative, the orientation is 'reverse'.
    //

    //
    // We need at least two orthologous genes to determine orientation.
    //
    if (genes.size() < 2)
        return ori_unknown;
    
    double n      = genes.size();
    double avg_x  = 0.0;
    double avg_y  = 0.0;
    double avg_xy = 0.0;
    double avg_x2 = 0.0;

    for (uint i = 0; i < genes.size(); i++) {
        // cerr << "orth bp: " << genes[i].orth_bp << " scaf bp: " << genes[i].bp << "\n";
        avg_x  += (double) genes[i].orth_bp;
        avg_y  += (double) genes[i].bp;
        avg_x2 += pow(genes[i].orth_bp, 2);
        avg_xy += genes[i].orth_bp * (double) genes[i].bp;
    }

    avg_x  = avg_x  / n;
    avg_y  = avg_y  / n;
    avg_x2 = avg_x2 / n;
    avg_xy = avg_xy / n;

    double m = ((avg_x * avg_y) - avg_xy) / (pow(avg_x, 2) - avg_x2);

    fh << "    Linear regression of orthologous genes versus local genes, slope (m): " << m << "\n";
    
    if (m < 0) 
        return ori_reverse;
    else if (m > 0)
        return ori_forward;

    return ori_unknown;
}

int
prune_outoforder_synteny_genes(ostream &doc_fh, ori_type scaffold_ori, vector<Ortholog> &genes)
{
    char doc_str[max_len];

    if (scaffold_ori == ori_unknown)
        return 0;

    doc_fh << "  Unmodified gene order, sorted according to ortholgous genome positions:\n";
    for (uint i = 0; i < genes.size(); i++) {
        snprintf(doc_str, max_len,
                 "    %03u\t%s\t%9lubp <=> %s\t%9lubp\n",
                 i, genes[i].orth_chr.c_str(), genes[i].orth_bp, genes[i].scaffold.c_str(), genes[i].bp);
        doc_fh << doc_str;
    }

    //
    // Examine each coherent path through the sorted markers (markers consistently down or up ordered).
    //
    int gene_index, cur, next, cnt;
    int stop           = (int) genes.size() - 1;
    int max_gene_cnt   = 0;
    int max_gene_index = 0;

    vector<int> path(genes.size(), 0);
    vector<int> blacklist(genes.size(), 0);
    vector<int> max_path;

    for (gene_index = 0; gene_index < stop; gene_index++) {
        for (cur = 0; cur < (int) path.size(); cur++) { path[cur] = 0; }

        cur       = gene_index;
        path[cur] = 1;
        cnt       = 1;

        if (scaffold_ori == ori_forward) {
            do {
                next = cur + 1;
                //
                // Skip out of order genes.
                //
                while (genes[next].bp < genes[cur].bp && next <= stop) next++;
                if (next <= stop) {
                    path[next] = 1;
                    cnt++;
                }
                cur = next;
            } while (cur < stop);

        } else {
            do {
                next = cur + 1;
                //
                // Skip out of order genes.
                //
                while (genes[next].bp > genes[cur].bp && next <= stop) next++;
                if (next <= stop) {
                    cnt++;
                    path[next] = 1;
                }
                cur = next;
            } while (cur < stop);
        }

        if (cnt > max_gene_cnt) {
            max_gene_cnt   = cnt;
            max_gene_index = gene_index;
            max_path         = path;
        }
    }

    // cerr << "    Max path: ";
    // for (uint j = 0; j < max_path.size(); j++)
    //     cerr << max_path[j];
    // cerr << "\n";

    doc_fh << "\n" 
           << "  Removing out of order genes:\n";

    //
    // Test to see if we can add genes to the maximal path.
    //
    for (int i = 0; i < max_gene_index; i++)
        blacklist[i] = 1;

    int index = (int) path.size() - 1;
    do {
        path = max_path;
        while (index > 0 && blacklist[index] == 0 && path[index] == 0) index--;
        //
        // Test this position to see if removing this marker adds more genes overall.
        //
        blacklist[index] = 1;
        cnt = genes_in_path(genes, path, blacklist, scaffold_ori);

        //
        // If it didn't add more genes, take it back out.
        //
        if (cnt <= max_gene_cnt) {
            blacklist[index] = 0;
        } else {
            max_gene_cnt = cnt;
            max_path       = path;
        }

        index--;
    } while (index > max_gene_index);

    index = 0;
    do {
        //
        // Test this position to see if adding this marker adds more genes overall.
        //
        blacklist[index] = 0;
        cnt = genes_in_path(genes, path, blacklist, scaffold_ori);

        //
        // If it didn't add more genes, take it back out.
        //
        if (cnt <= max_gene_cnt) {
            blacklist[index] = 1;
        } else {
            max_gene_cnt = cnt;
            max_path     = path;
        }

        index++;
    } while (index < max_gene_index);
	    
    vector<Ortholog> tmp;
    for (uint i = 0; i < genes.size(); i++)
        tmp.push_back(genes[i]);
    genes.clear();

    //
    // Discard genes that are not on the final marker path.
    //
    uint removed_cnt = 0;
    for (uint i = 0; i < max_path.size(); i++)
        if (max_path[i]) {
            genes.push_back(tmp[i]);
        } else {
            snprintf(doc_str, max_len, "  %03u: removing gene %s [%8lubp]\n", i, tmp[i].id.c_str(), tmp[i].bp);
            doc_fh << doc_str;
            removed_cnt++;
        }

    if (removed_cnt == 0)
        doc_fh << "    No genes were removed.\n";
    
    doc_fh << "\n"
           << "  Final conserved synteny gene ordering:\n";

    for (uint i = 0; i < genes.size(); i++) {
        snprintf(doc_str, max_len,
                 "    %03u\t%s\t%9lubp <=> %s\t%9lubp\n",
                 i, genes[i].orth_chr.c_str(), genes[i].orth_bp, genes[i].scaffold.c_str(), genes[i].bp);
        doc_fh << doc_str;
    }
    doc_fh << "\n";

    return 0;
}

int
genes_in_path(vector<Ortholog> &genes, vector<int> &path, vector<int> &blacklist, ori_type orientation)
{
    for (uint i = 0; i < path.size(); i++) path[i] = 0;

    uint next;
    uint stop = genes.size() - 1;
    uint cur  = 0;
    uint cnt  = 1;

    while (cur <= stop && blacklist[cur] == 1)
	cur++;
    path[cur] = 1;

    if (orientation == ori_forward) {
	do {
	    next = cur + 1;
	    //
	    // Skip out of order markers.
	    //
	    while (blacklist[next] == 1 || (genes[next].bp < genes[cur].bp && next <= stop)) next++;
	    if (next <= stop) {
		path[next] = 1;
		cnt++;
	    }
	    cur = next;
	} while (cur < stop);

    } else {
	do {
	    next = cur + 1;
	    //
	    // Skip out of order markers.
	    //
	    while (blacklist[next] == 1 || (genes[next].bp > genes[cur].bp && next <= stop)) next++;
	    if (next <= stop) {
		cnt++;
		path[next] = 1;
	    }
	    cur = next;
	} while (cur < stop);
    }

    return cnt;
}

int
parse_gtf_file(string path, vector<GeneAnn> &annotations)
{
    char *line      = new char[max_len];
    int   size      = max_len;
    int   fh_status = 1;
    int   line_num  = 1;

    vector<string> parts, attributes;

    ifstream fh;
    gzFile   gzfh;
    bool compressed = false;

    uint pos = path.find_last_of(".");
    
    if (path.substr(pos + 1) == "gz")
        compressed = true;

    if (compressed) {
        gzfh = gzopen(path.c_str(), "rb");
        if (!gzfh) {
            cerr << "Unable to open gzipped file '" << path << "': " << strerror(errno) << ".\n";
            return 0;
        }
    } else {
        fh.open(path.c_str(), ifstream::in);
        if (fh.fail()) {
            cerr << " Unable to open '" << path << "'\n";
            return 0;
        }
    }
    cerr << "  Parsing '" << path << "'.\n";

    while (fh_status) {
        if (compressed)
            fh_status = read_gzip_line(gzfh, &line, &size);
        else
            fh_status = read_line(fh, &line, &size);

        if (!fh_status && strlen(line) == 0)
	    continue;

	if (is_comment(line)) continue;

	parse_tsv(line, parts);

        if (parts.size() != num_gtf_fields) {
            cerr << "Error parsing " << path << " at line: " << line_num << ". (" << parts.size() << " fields).\n";
            return 0;
        }

        line_num++;

        //
        // Ignore all the elements except genes.
        //
        if (parts[2] != "gene")
            continue;
        
        string    scaffold = parts[0];
        size_t    bp       = is_integer(parts[3].c_str());

        //
        // Parse attributes in column 9 to find "gene_id".
        //
        parse_delimiter(parts[8].c_str(), ';', attributes);

        if (attributes.size() == 0) {
            cerr << "Error parsing attributes (column 9) of " << path << " at line: " << line_num << ".\n";
            return 0;
        }
        
        string gene_id = "";
        for (uint i = 0; i < attributes.size(); i++) {
            pos = attributes[i].find_first_of(' ');
            if (attributes[i].substr(0,pos) != "gene_id")
                continue;
            gene_id = attributes[i].substr(pos+1);

            if ( (gene_id[0] == '"' || gene_id[0] == '\'') &&
                 (gene_id[gene_id.length() - 1] == '"' || gene_id[gene_id.length() - 1] == '\''))
                gene_id = gene_id.substr(1, gene_id.length() - 2);
        }

        if (gene_id.length() == 0) {
            cerr << "Error parsing gene ID (column 9) of " << path << " at line: " << line_num << ".\n";
            return 0;
        }
        
        annotations.push_back(GeneAnn(gene_id, scaffold, bp));
    }

    if (compressed)
        gzclose(gzfh);
    else
        fh.close();

    delete [] line;

    cerr << "    Loaded " << annotations.size() << " genes.\n";

    return 1;
}

int
parse_gff_file(string path, vector<GeneAnn> &annotations)
{
    char *line      = new char[max_len];
    int   size      = max_len;
    int   fh_status = 1;
    int   line_num  = 1;

    vector<string> parts, attributes;

    ifstream fh;
    gzFile   gzfh;
    bool compressed = false;

    uint pos = path.find_last_of(".");
    
    if (path.substr(pos + 1) == "gz")
        compressed = true;

    if (compressed) {
        gzfh = gzopen(path.c_str(), "rb");
        if (!gzfh) {
            cerr << "Unable to open gzipped file '" << path << "': " << strerror(errno) << ".\n";
            return 0;
        }
    } else {
        fh.open(path.c_str(), ifstream::in);
        if (fh.fail()) {
            cerr << " Unable to open '" << path << "'\n";
            return 0;
        }
    }
    cerr << "  Parsing '" << path << "'.\n";

    while (fh_status) {
        if (compressed)
            fh_status = read_gzip_line(gzfh, &line, &size);
        else
            fh_status = read_line(fh, &line, &size);

        if (!fh_status && strlen(line) == 0)
	    continue;

	if (is_comment(line)) continue;

	parse_tsv(line, parts);

        if (parts.size() != num_gtf_fields) {
            cerr << "Error parsing " << path << " at line: " << line_num << ". (" << parts.size() << " fields).\n";
            return 0;
        }

        line_num++;

        //
        // Ignore all the elements except genes.
        //
        if (parts[2] != "gene")
            continue;
        
        string    scaffold = parts[0];
        size_t    bp       = is_integer(parts[3].c_str());

        //
        // Parse attributes in column 9 to find "gene_id".
        //
        parse_delimiter(parts[8].c_str(), ';', attributes);

        if (attributes.size() == 0) {
            cerr << "Error parsing attributes (column 9) of " << path << " at line: " << line_num << ".\n";
            return 0;
        }
        
        string gene_id = "";
        for (uint i = 0; i < attributes.size(); i++) {
            pos = attributes[i].find_first_of('=');
            if (attributes[i].substr(0,pos) != "ID")
                continue;
            gene_id = attributes[i].substr(pos+1);

            if ( (gene_id[0] == '"' || gene_id[0] == '\'') &&
                 (gene_id[gene_id.length() - 1] == '"' || gene_id[gene_id.length() - 1] == '\''))
                gene_id = gene_id.substr(1, gene_id.length() - 2);
        }

        if (gene_id.length() == 0) {
            cerr << "Error parsing gene ID (column 9) of " << path << " at line: " << line_num << ".\n";
            return 0;
        }
        
        annotations.push_back(GeneAnn(gene_id, scaffold, bp));
    }

    if (compressed)
        gzclose(gzfh);
    else
        fh.close();

    delete [] line;

    cerr << "    Loaded " << annotations.size() << " genes.\n";

    return 1;
}

int
parse_orthologs(string path, map<string, Ortholog> &orthologs, map<string, string> &ortholog_key)
{
    char *line      = new char[max_len];
    int   size      = max_len;
    int   fh_status = 1;
    int   line_num  = 1;

    vector<string> parts, attributes;

    ifstream fh;
    gzFile   gzfh;
    bool compressed = false;

    uint pos = path.find_last_of(".");
    
    if (path.substr(pos + 1) == "gz")
        compressed = true;

    if (compressed) {
        gzfh = gzopen(path.c_str(), "rb");
        if (!gzfh) {
            cerr << "Unable to open gzipped file '" << path << "': " << strerror(errno) << ".\n";
            return 0;
        }
    } else {
        fh.open(path.c_str(), ifstream::in);
        if (fh.fail()) {
            cerr << " Unable to open '" << path << "'\n";
            return 0;
        }
    }
    cerr << "  Parsing '" << path << "'.\n";

    while (fh_status) {
        if (compressed)
            fh_status = read_gzip_line(gzfh, &line, &size);
        else
            fh_status = read_line(fh, &line, &size);

        if (!fh_status && strlen(line) == 0)
	    continue;

	if (is_comment(line)) continue;

	parse_tsv(line, parts);

        if (parts.size() != 2) {
            cerr << "Error parsing ortholog file '" << path << "' at line: " << line_num << ". (" << parts.size() << " fields, expected two).\n";
            return 0;
        }

        line_num++;

        orthologs[parts[0]]    = Ortholog(parts[0], parts[1]);
        ortholog_key[parts[1]] = parts[0];
    }

    if (compressed)
        gzclose(gzfh);
    else
        fh.close();

    delete [] line;

    cerr << "    Loaded " << orthologs.size() << " ortholog gene pairs.\n";

    return 1;
}

int
process_sort_orthologs(map<string, Ortholog> &orthologs,
                       map<string, string> &ortholog_key,
                       vector<GeneAnn> &gene_ann,
                       vector<GeneAnn> &orth_ann,
                       map<string, vector<Ortholog>> &con_synteny)
{
    size_t gene_ann_found = 0;
    size_t orth_ann_found = 0;

    for (uint i = 0; i < gene_ann.size(); i++) {
        if (orthologs.count(gene_ann[i].id) == 0)
            continue;

        orthologs[gene_ann[i].id].scaffold = gene_ann[i].scaffold;
        orthologs[gene_ann[i].id].bp       = gene_ann[i].bp;
        gene_ann_found++;
    }

    for (uint i = 0; i < orth_ann.size(); i++) {
        if (ortholog_key.count(orth_ann[i].id) == 0)
            continue;

        string g = ortholog_key[orth_ann[i].id];
        
        orthologs[g].orth_chr = orth_ann[i].scaffold;
        orthologs[g].orth_bp  = orth_ann[i].bp;
        orth_ann_found++;
    }

    //
    // Categorize the orthologs by scaffold and sort them by basepair.
    //
    for (auto it = orthologs.begin(); it != orthologs.end(); it++)
        if (it->second.scaffold.length() > 0)
            con_synteny[it->second.scaffold].push_back(it->second);

    size_t cnt = 0;
    for (auto it = con_synteny.begin(); it != con_synteny.end(); it++) {
        cnt += it->second.size();
        sort(it->second.begin(), it->second.end(), [](Ortholog a, Ortholog b) { return a.bp > b.bp; });
    }

    cerr << "  Of the " << orthologs.size() << " gene pairs specified: " << gene_ann_found << " were found in the gene annotation; " << orth_ann_found << " were found in the ortholog annotation.\n"
         << "    " << cnt << " ortholog gene pairs, located on " << con_synteny.size() << " scaffolds, available for conserved synteny analysis.\n";

    return 0;
}
